package main

import (
	"fmt"
	"os"

	log "github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"go.mills.io/saltyim"
	"go.mills.io/saltyim/internal/tui"
)

var chatCmd = &cobra.Command{
	Use:     "chat <user>",
	Aliases: []string{"talk"},
	Short:   "Creates a chat with a specific user",
	Long: `This command creates a chat with the specified user by discovering
and subscribing to your endpoint and prompts for input and sends encrypted
messages to the user via their discovered endpoint.`,
	Args: cobra.MinimumNArgs(1),
	Run: func(cmd *cobra.Command, args []string) {
		user := viper.GetString("user")
		identity := viper.GetString("identity")
		state := viper.GetString("state")

		var profiles []profile
		viper.UnmarshalKey("profiles", &profiles)
		for _, p := range profiles {
			if user == p.User {
				identity = p.Identity
			}
		}

		me, err := saltyim.ParseAddr(user)
		if err != nil {
			fmt.Fprintf(os.Stderr, "error parsing addr: %s\n", err)
		}
		// XXX: What if me.IsZero()

		chat(me, identity, state, args[0])
	},
}

func init() {
	rootCmd.AddCommand(chatCmd)
}

func chat(me *saltyim.Addr, identity, state, user string) {
	cli, err := saltyim.NewClient(me,
		saltyim.WithStateFromFile(state),
		saltyim.WithClientIdentity(saltyim.WithIdentityPath(identity)),
	)
	if err != nil {
		fmt.Fprintf(os.Stderr, "error initializing client: %s\n", err)
		os.Exit(2)
	}
	defer func() {
		if err := cli.State().Save(state); err != nil {
			log.WithError(err).Warnf("error saving state: %s", state)
		}
	}()

	// Set terminal title
	tui.SetTerminalTitle("Salty IM with %s", user)

	// Initialize necessary channels.
	inCh := make(chan string)
	outCh := make(chan string)

	// Initialize ui.
	ui, err := tui.NewChatTUI(cli, user)
	if err != nil {
		fmt.Fprintf(os.Stderr, "error creating chat: %s\n", err)
		os.Exit(2)
	}

	// Run the ui loop
	go ui.RunChat(inCh, outCh)
	ui.SetScreen(inCh, outCh)
}
