package main

import (
	"fmt"
	"os"
	"os/signal"
	"syscall"

	"github.com/mattn/go-isatty"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"golang.org/x/net/context"

	"go.mills.io/saltyim"
)

var readCmd = &cobra.Command{
	Use:   "read [<inbox>]",
	Short: "Reads messages from an inbox",
	Long: `This command subscribes to the provided inbox (optiona) which if
not specified defaults to the local user ($USER)`,
	Run: func(cmd *cobra.Command, args []string) {
		user := viper.GetString("user")
		identity := viper.GetString("identity")
		state := viper.GetString("state")

		var profiles []profile
		viper.UnmarshalKey("profiles", &profiles)
		for _, p := range profiles {
			if user == p.User {
				identity = p.Identity
			}
		}

		me, err := saltyim.ParseAddr(user)
		if err != nil {
			fmt.Fprintf(os.Stderr, "error parsing addr: %s\n", err)
		}
		// XXX: What if me.IsZero()

		follow, err := cmd.Flags().GetBool("follow")
		if err != nil {
			log.Fatal("error getting -f--follow flag")
		}

		extraenvs, err := cmd.Flags().GetString("extra-envs")
		if err != nil {
			log.Fatal("error getting --extra-envs flag")
		}

		prehook, err := cmd.Flags().GetString("pre-hook")
		if err != nil {
			log.Fatal("error getting --pre-hook flag")
		}

		posthook, err := cmd.Flags().GetString("post-hook")
		if err != nil {
			log.Fatal("error getting --post-hook flag")
		}

		read(me, identity, state, follow, extraenvs, prehook, posthook, args...)
	},
}

type profile struct {
	User     string
	Identity string
}

func init() {
	rootCmd.AddCommand(readCmd)

	readCmd.Flags().BoolP(
		"follow", "f", false,
		"Subscribe to the inbox and follow all messages",
	)

	readCmd.Flags().String(
		"extra-envs", "",
		"List of extra env vars to pass to pre/post hooks (KEY=[VALUE] ...)",
	)

	readCmd.Flags().String(
		"pre-hook", "",
		"Execute pre-hook before message decryption",
	)

	readCmd.Flags().String(
		"post-hook", "",
		"Execute post-hook after message decryption",
	)
}

func read(me *saltyim.Addr, identity, state string, follow bool, extraenvs, prehook, posthook string, args ...string) {
	cli, err := saltyim.NewClient(me,
		saltyim.WithStateFromFile(state),
		saltyim.WithClientIdentity(saltyim.WithIdentityPath(identity)),
	)
	if err != nil {
		fmt.Fprintf(os.Stderr, "error initializing client: %s\n", err)
		os.Exit(2)
	}
	defer func() {
		if err := cli.State().Save(state); err != nil {
			log.WithError(err).Warnf("error saving state: %s", state)
		}
	}()

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM)

	go func() {
		<-sigs
		cancel()
	}()

	if follow {
		for msg := range cli.Subscribe(ctx, extraenvs, prehook, posthook) {
			if isatty.IsTerminal(os.Stdout.Fd()) {
				fmt.Println(saltyim.FormatMessage(msg.Text))
			} else {
				fmt.Println(msg.Text)
			}
		}
	} else {
		msg, err := cli.Read(extraenvs, prehook, posthook)
		if err != nil {
			if err == saltyim.ErrNoMessages {
				os.Exit(0)
			}
			fmt.Fprintf(os.Stderr, "error reading message: %s\n", err)
			os.Exit(2)
		}
		if isatty.IsTerminal(os.Stdout.Fd()) {
			fmt.Println(saltyim.FormatMessage(msg.Text))
		} else {
			fmt.Println(msg.Text)
		}
	}
}
